
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

static void	Error(head,tail)
char		*head,*tail;
{
	(void)fprintf(stderr,"%s : %s\n",head,tail); exit(1);
}

static char	*Dir;

static char	**AllocA(names)
int		names;
{
	char	**ptr=(char **)malloc(sizeof(*ptr)*(size_t)names);

	if (ptr==NULL) Error(Dir,"name array memory allocation error.");

	return ptr;
}

static char	*AllocS(name)
char		*name;
{
	char	*ptr=strdup(name);

	if (ptr==NULL) Error(Dir,"name string memory allocation error.");

	return ptr;
}

static int	Normal(name1,name2)
char		**name1,**name2;
{
	return strcmp(*name1,*name2);
}

static int	Reverse(name1,name2)
char		**name1,**name2;
{
	return Normal(name2,name1);
}

static int	Numeric(name,val)
char		*name;
double		*val;
{
	do
	    if (*name=='-' || *name=='+' ||
	       (*name>='0' && *name<='9')) return sscanf(name,"%lf",val)==1;
	while (*(++name)!='\0');

	return 0;
}

static int	Ascending(name1,name2)
char		**name1,**name2;
{
	double	val1,val2;
	int	flag1=Numeric(*name1,&val1),flag2=Numeric(*name2,&val2);

	return (flag1 && flag2)?(val1>val2)?1:(val1<val2)?-1:0:
	       (flag1)?1:(flag2)?-1:Normal(name1,name2);
}

static int	Descending(name1,name2)
char		**name1,**name2;
{
	return Ascending(name2,name1);
}

#define PATH_LEN	1024

static char	path[PATH_LEN];

static int	FGetC(file)
FILE		*file;
{
	int	code;

	if ((code=fgetc(file))==EOF) Error(path,"unexpected end of file.");

	return code;
}

static void	Skip(file,bytes)
FILE		*file;
int		bytes;
{
	while (--bytes>=0) (void)FGetC(file);
}

static void	Read(file,size,cm,rgb)
FILE		*file;
int		size;
unsigned char	*cm;
double		rgb;
{
	if (fread(cm,(size_t)size,(size_t)1,file)!=1)
	    Error(path,"colormap read error.");

	while (--size>=0) cm[size]=(unsigned char)((double)cm[size]*rgb+0.5);
}

#define MEMCMP(str1,str2,size)	memcmp(str1,str2,(size_t)(size))

static void	Open(name,file,rgb,bps,read,cm)
char		*name;
FILE		**file;
double		*rgb;
int		*bps,read;
unsigned char	*cm;
{
	char	magic[6];
	int	flag,idx;

	(void)sprintf(path,"%s/%s",Dir,name);

	if ((*file=fopen(path,"rb"))==NULL) Error(path,"GIF file not open.");

	if (fread(magic,(size_t)6,(size_t)1,*file)!=1 ||
	   (MEMCMP("GIF87a",magic,6) &&
	    MEMCMP("GIF89a",magic,6))) Error(path,"not a GIF file.");

	Skip(*file,4);
	flag=FGetC(*file);
	Skip(*file,2);

	*rgb=255.0/(double)((1<<((flag>>4)&7)+1)-1);

	*bps=(flag&7)+1;
	if (flag&128)
	    if (read) {
		idx=(1<<*bps)*3;
		Read(*file,idx,cm,*rgb);
		while (idx<256*3) cm[idx++]=0;
	    }
	    else
		Skip(*file,(1<<*bps)*3);
	else
	    if (read)
		for (idx=0; idx<256*3; idx++) cm[idx]=(unsigned char)(idx/3);
}

static int	FGetW(file)
FILE		*file;
{
	int	code=FGetC(file);

	return code|FGetC(file)<<8;
}

#define FPUTC	(void)fputc

static void	FPutW(word,file)
int		word;
FILE		*file;
{
	FPUTC(word&255,file); FPUTC(word>>8,file);
}

#define FWRITE(ptr,size,noe,file)	\
	(void)fwrite(ptr,(size_t)(size),(size_t)(noe),file)

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE		*file,*fileI,*fileO;
	int		No,no,ei,bpsG,nx,ny,flag,bpsL,size,
			Nx=0,Ny=0,same=1,bps=0;
	char		str[PATH_LEN],**name,*top,
			*nifn="no image file name.",
			*bnoifn="bad number of image file name.",
			*uei="unknown extension introducer.";
	DIR		*dir;
	struct dirent	*ent;
	double		rgb;
	unsigned long	dt,rc;
	unsigned char	cmG[256*3],cmL[256*3],cm[256*3];

	if (argc!=6)
	Error("usage","gif_movie directory nameFile delayTime repeatCount GIF");

	Dir=argv[1];

#define NF	argv[2]

	if (NF[0]!='-') {
	    if ((file=fopen(NF,"r"))==NULL) Error(NF,"name file not open.");

	    for (No=0; fscanf(file,"%s",str)==1; No++) ;

	    if (No==0) Error(NF,nifn);

	    name=AllocA(No); rewind(file);

	    for (no=0; fscanf(file,"%s",str)==1; no++)
		if (no<No) name[no]=AllocS(str);

	    if (no!=No) Error(NF,bnoifn);

	    (void)fclose(file);
	}
	else if (NF[1]!='\0' && NF[1]!='-') {
	    top=NF+1;
	    for (No=0; sscanf(top,"%s",str)==1; No++)
		top=strstr(top,str)+strlen(str);

	    if (No==0) Error(NF+1,nifn);

	    name=AllocA(No);

	    top=NF+1;
	    for (no=0; sscanf(top,"%s",str)==1; no++) {
		if (no<No) name[no]=AllocS(str);

		top=strstr(top,str)+strlen(str);
	    }
	    if (no!=No) Error(NF+1,bnoifn);
	}
	else {
	    if ((dir=opendir(Dir))==NULL) Error(Dir,"directory not open.");

	    for (No=0; (ent=readdir(dir))!=NULL; )
		if (strcmp(ent->d_name,".") && strcmp(ent->d_name,"..")) ++No;

	    if (No==0) Error(Dir,nifn);

	    name=AllocA(No); rewinddir(dir);

	    for (no=0; (ent=readdir(dir))!=NULL; )
		if (strcmp(ent->d_name,".") && strcmp(ent->d_name,".."))
		    if (++no<=No) name[no-1]=AllocS(ent->d_name);

	    if (no!=No) Error(Dir,bnoifn);

	    (void)closedir(dir);

	    if (NF[1]=='\0')
		qsort(name,(size_t)No,sizeof(*name),Normal);
	    else if (NF[2]=='r' && NF[3]=='\0')
		qsort(name,(size_t)No,sizeof(*name),Reverse);
	    else if (NF[2]=='n' && NF[3]=='\0')
		qsort(name,(size_t)No,sizeof(*name),Ascending);
	    else if ((NF[2]=='r' && NF[3]=='n') ||
		     (NF[2]=='n' && NF[3]=='r'))
		qsort(name,(size_t)No,sizeof(*name),Descending);
	    else
		Error(NF+1,"unkown sorting option.");
	}
	if ((dt=atol(argv[3]))>65535) Error(argv[3],"bad delay time.");
	if ((rc=atol(argv[4]))>65536) Error(argv[4],"bad repeat count.");

	for (no=0; no<No; no++) {
	    Open(name[no],&fileI,&rgb,&bpsG,same,cmG);

	    while ((ei=FGetC(fileI))!=';')
		if (ei==',' || ei=='!') {
		    if (ei==',') {
			Skip(fileI,4);
			if ((nx=FGetW(fileI))>Nx) Nx=nx;
			if ((ny=FGetW(fileI))>Ny) Ny=ny;

			flag=FGetC(fileI); bpsL=(flag&7)+1; size=(1<<bpsL)*3;
			if (flag&128)
			    if (same) {
				Read(fileI,size,cmL,rgb);
				if (bps==0)
				    (void)memcpy(cm,cmL,(size_t)size);
				else
				    same=bps==bpsL && !MEMCMP(cm,cmL,size);
			    }
			    else
				Skip(fileI,size);
			else
			    if (same)
				if (bps==0)
				    (void)memcpy(cm,cmG,(size_t)size);
				else
				    same=bps==bpsL && !MEMCMP(cm,cmG,size);

			if (bpsL>bps) bps=bpsL;
		    }
		    Skip(fileI,1);
		    while ((size=FGetC(fileI))!=0) Skip(fileI,size);
		}
		else
		    Error(path,uei);

	    (void)fclose(fileI);
	}
	if (bps==0)
	    Error(strcat(strcat(strcpy(path,Dir)," "),NF),"no image data.");

#define GIF	argv[5]

	if (GIF[0]=='-')
	    fileO=stdout;
	else
	    if ((fileO=fopen(GIF,"wb"))==NULL) Error(GIF,"GIF file not open.");

	FWRITE((dt==0 && rc==0)?"GIF87a":"GIF89a",6,1,fileO);
	FPutW(Nx,fileO);
	FPutW(Ny,fileO);
	FPUTC(((same)?128:0)|112|(bps-1),fileO);
	FPUTC(0,fileO);
	FPUTC(0,fileO);
	if (same) FWRITE(cm,(1<<bps)*3,1,fileO);

	if (rc!=0) {
	    FPUTC('!',fileO); FPUTC(255,fileO);
	    FPUTC(11,fileO); FWRITE("NETSCAPE2.0",11,1,fileO);
	    FPUTC(3,fileO); FPUTC(1,fileO); FPutW((int)(rc&65535),fileO);
	    FPUTC(0,fileO);
	}
	for (no=0; no<No; no++) {
	    Open(name[no],&fileI,&rgb,&bpsG,!same,cmG);

	    while ((ei=FGetC(fileI))!=';')
		if (ei==',') {
		    Skip(fileI,4);
		    if ((nx=FGetW(fileI))>Nx ||
			(ny=FGetW(fileI))>Ny) Error(path,"bad image size.");

		    FPUTC(',',fileO);
		    FPutW((Nx-nx)/2,fileO);
		    FPutW((Ny-ny)/2,fileO);
		    FPutW(nx,fileO);
		    FPutW(ny,fileO);

		    flag=FGetC(fileI);

		    bpsL=(flag&7)+1;
		    if (bpsL>bps) Error(path,"bad bits per sample.");

		    size=(1<<bpsL)*3;
		    if (same) {
			FPUTC(flag&71,fileO);

			if (flag&128) Skip(fileI,size);
		    }
		    else {
			FPUTC(flag&223,fileO);

			if (flag&128) {
			    Read(fileI,size,cmL,rgb); FWRITE(cmL,size,1,fileO);
			}
			else
			    FWRITE(cmG,size,1,fileO);
		    }
		    FPUTC(FGetC(fileI),fileO);
		    while ((size=FGetC(fileI))!=0) {
			FPUTC(size,fileO);
			while (--size>=0) FPUTC(FGetC(fileI),fileO);
		    }
		    FPUTC(0,fileO);

		    if (dt!=0) {
			FPUTC('!',fileO); FPUTC(249,fileO);
	FPUTC(4,fileO); FPUTC(0,fileO); FPutW((int)dt,fileO); FPUTC(0,fileO);
			FPUTC(0,fileO);
		    }
		}
		else if (ei=='!') {
		    Skip(fileI,1);
		    while ((size=FGetC(fileI))!=0) Skip(fileI,size);
		}
		else
		    Error(path,uei);

	    (void)fclose(fileI);
	}
	FPUTC(';',fileO); if (fileO!=stdout) (void)fclose(fileO);

	return 0;
}
